#     Optopsy - Python Backtesting library for options trading strategies
#     Copyright (C) 2018  Michael Chu

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .enums import OptionType, OrderAction
from .backtest import create_spread
from .checks import (
    singles_checks,
    call_spread_checks,
    put_spread_checks,
    iron_condor_checks,
    iron_condor_spread_check,
)
from datetime import datetime


def _process_legs(data, legs, filters, check_func, sort_by=None, asc=True):
    if _filter_checks(filters, check_func):
        return create_spread(data, legs, filters, sort_by=sort_by, ascending=asc)
    else:
        raise ValueError(
            "Invalid filter values provided, please check the filters and try again."
        )


def _filter_checks(filter, func=None):
    return True if func is None else func(filter)


def _merge(filters, start, end):
    return {**filters, **{"start_date": start, "end_date": end}}


def long_call(data, start, end, filters):
    legs = [(OptionType.CALL, 1)]
    return _process_legs(data, legs, _merge(filters, start, end), singles_checks)


def short_call(data, start, end, filters):
    legs = [(OptionType.CALL, -1)]
    return _process_legs(data, legs, _merge(filters, start, end), singles_checks)


def long_put(data, start, end, filters):
    legs = [(OptionType.PUT, 1)]
    return _process_legs(data, legs, _merge(filters, start, end), singles_checks)


def short_put(data, start, end, filters):
    legs = [(OptionType.PUT, -1)]
    return _process_legs(data, legs, _merge(filters, start, end), singles_checks)


def long_call_spread(data, start, end, filters):
    legs = [(OptionType.CALL, 1), (OptionType.CALL, -1)]
    return _process_legs(data, legs, _merge(filters, start, end), call_spread_checks)


def short_call_spread(data, start, end, filters):
    legs = [(OptionType.CALL, -1), (OptionType.CALL, 1)]
    return _process_legs(data, legs, _merge(filters, start, end), call_spread_checks)


def long_put_spread(data, start, end, filters):
    legs = [(OptionType.PUT, -1), (OptionType.PUT, 1)]
    return _process_legs(data, legs, _merge(filters, start, end), put_spread_checks)


def short_put_spread(data, start, end, filters):
    legs = [(OptionType.PUT, 1), (OptionType.PUT, -1)]
    return _process_legs(data, legs, _merge(filters, start, end), put_spread_checks)


def long_iron_condor(data, start, end, filters):
    legs = [
        (OptionType.PUT, 1),
        (OptionType.PUT, -1),
        (OptionType.CALL, -1),
        (OptionType.CALL, 1),
    ]
    return _process_legs(
        data, legs, _merge(filters, start, end), iron_condor_checks
    ).pipe(iron_condor_spread_check)


def short_iron_condor(data, start, end, filters):
    legs = [
        (OptionType.PUT, -1),
        (OptionType.PUT, 1),
        (OptionType.CALL, 1),
        (OptionType.CALL, -1),
    ]
    return _process_legs(
        data, legs, _merge(filters, start, end), iron_condor_checks
    ).pipe(iron_condor_spread_check)



#     Optopsy - Python Backtesting library for options trading strategies
#     Copyright (C) 2018  Michael Chu

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np


def _calc_opt_px(data, action):
    ask = f"ask_{action}"
    bid = f"bid_{action}"

    if action == "entry":
        return np.where(data["ratio"] > 0, data[ask], data[bid])
    elif action == "exit":
        return np.where(data["ratio"] > 0, data[bid], data[ask])
    return data


def _assign_opt_px(data, mode, action):
    if mode == "midpoint":
        bid_ask = [f"bid_{action}", f"ask_{action}"]
        data[f"{action}_opt_price"] = data[bid_ask].mean(axis=1)
    elif mode == "market":
        data[f"{action}_opt_price"] = _calc_opt_px(data, action)
    return data


def assign_trade_num(data, groupby):
    data["trade_num"] = data.groupby(groupby).ngroup()
    data.set_index("trade_num", inplace=True)
    return data


def calc_entry_px(data, mode="midpoint"):
    return _assign_opt_px(data, mode, "entry")


def calc_exit_px(data, mode="midpoint"):
    return _assign_opt_px(data, mode, "exit")


def calc_pnl(data):
    # calculate the p/l for the trades
    data["entry_price"] = (
        data["entry_opt_price"] * data["ratio"] * data["contracts"] * 100
    )
    data["exit_price"] = (
        data["exit_opt_price"] * data["ratio"] * -1 * data["contracts"] * 100
    )
    data["cost"] = data["exit_price"] + data["entry_price"]
    return data.round(2)


def calc_total_profit(data):
    return data["cost"].sum().round(2)



import datetime

filters = {
    "start_date": datetime.date,
    "end_date": datetime.date,
    "std_expr": bool,
    "contract_size": int,
    "entry_dte": (int, tuple),
    "entry_days": int,
    "leg1_delta": (int, float, tuple),
    "leg2_delta": (int, float, tuple),
    "leg3_delta": (int, float, tuple),
    "leg4_delta": (int, float, tuple),
    "leg1_strike_pct": (int, float, tuple),
    "leg2_strike_pct": (int, float, tuple),
    "leg3_strike_pct": (int, float, tuple),
    "leg4_strike_pct": (int, float, tuple),
    "entry_spread_price": (int, float, tuple),
    "entry_spread_delta": (int, float, tuple),
    "entry_spread_yield": (int, float, tuple),
    "exit_dte": int,
    "exit_hold_days": int,
    "exit_leg_1_delta": (int, float, tuple),
    "exit_leg_1_otm_pct": (int, float, tuple),
    "exit_profit_loss_pct": (int, float, tuple),
    "exit_spread_delta": (int, float, tuple),
    "exit_spread_price": (int, float, tuple),
    "exit_strike_diff_pct": (int, float, tuple),
}


def _type_check(filter):
    return all([isinstance(filter[f], filters[f]) for f in filter])


def singles_checks(filter):
    return "leg1_delta" in filter and _type_check(filter)


def _sanitize(filters, f):
    return filters[f][1] if isinstance(filters[f], tuple) else filters[f]


def call_spread_checks(f):
    return (
        "leg1_delta" in f
        and "leg2_delta" in f
        and _type_check(f)
        and (_sanitize(f, "leg1_delta") > _sanitize(f, "leg2_delta"))
    )


def put_spread_checks(f):
    return (
        "leg1_delta" in f
        and "leg2_delta" in f
        and _type_check(f)
        and (_sanitize(f, "leg1_delta") < _sanitize(f, "leg2_delta"))
    )


def iron_condor_checks(f):
    return (
        "leg1_delta" in f
        and "leg2_delta" in f
        and "leg3_delta" in f
        and "leg4_delta" in f
        and _type_check(f)
        and (_sanitize(f, "leg1_delta") < _sanitize(f, "leg2_delta"))
        and (_sanitize(f, "leg3_delta") > _sanitize(f, "leg4_delta"))
    )


def iron_condor_spread_check(ic):
    return (
        ic.assign(d_strike=lambda r: ic.duplicated(subset="strike", keep=False))
        .groupby(ic.index)
        .filter(lambda r: (r.d_strike == False).all())
        .drop(columns="d_strike")
    )



#     Optopsy - Python Backtesting library for options trading strategies
#     Copyright (C) 2018  Michael Chu

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from functools import reduce

import pandas as pd

import optopsy.filters as fil
from .option_queries import opt_type
from .statistics import *

pd.set_option("display.expand_frame_repr", False)


on = ["underlying_symbol", "option_type", "expiration", "strike"]

default_entry_filters = {
    "std_expr": False,
    "contract_size": 10,
    "entry_dte": (27, 30, 31),
    "exit_dte": None,
}

output_cols = {
    "quote_date_entry": "entry_date",
    "quote_date_exit": "exit_date",
    "delta_entry": "entry_delta",
    "underlying_price_entry": "entry_stk_price",
    "underlying_price_exit": "exit_stk_price",
    "dte_entry": "dte",
}

output_format = [
    "entry_date",
    "exit_date",
    "expiration",
    "underlying_symbol",
    "dte",
    "ratio",
    "contracts",
    "option_type",
    "strike",
    "entry_delta",
    "entry_stk_price",
    "exit_stk_price",
    "entry_opt_price",
    "exit_opt_price",
    "entry_price",
    "exit_price",
    "cost",
]


def _create_legs(data, leg):
    return data.pipe(opt_type, option_type=leg[0]).assign(ratio=leg[1])


def _apply_filters(legs, filters):
    if not filters:
        return legs
    else:
        return [
            reduce(lambda l, f: getattr(fil, f)(l, filters[f], idx), filters, leg)
            for idx, leg in enumerate(legs)
        ]


def _filter_data(data, filters):
    data = data if isinstance(data, list) else [data]
    return pd.concat(_apply_filters(data, filters))


def _do_dedupe(spread, groupby, col, mode):
    # dedupe delta dist ties
    if groupby is None:
        groupby = [
            "quote_date",
            "expiration",
            "underlying_symbol",
            "ratio",
            "option_type",
        ]

    on = groupby + [col]

    if mode == "min":
        return spread.groupby(groupby)[col].min().to_frame().merge(spread, on=on)
    else:
        return spread.groupby(groupby)[col].max().to_frame().merge(spread, on=on)


def _dedup_rows_by_cols(spreads, cols, groupby=None, mode="max"):
    return reduce(lambda i, c: _do_dedupe(spreads, groupby, c, mode), cols, spreads)


def create_spread(data, leg_structs, filters, sort_by, ascending):
    legs = [_create_legs(data, leg) for leg in leg_structs]
    sort_by = (
        ["quote_date", "expiration", "underlying_symbol", "strike"]
        if sort_by is None
        else sort_by
    )

    # merge and apply leg filters to create spread
    filters = {**default_entry_filters, **filters}
    entry_filters = {
        f: filters[f]
        for f in filters
        if (not f.startswith("entry_spread") and not f.startswith("exit_"))
    }
    spread = _filter_data(legs, entry_filters)

    # apply spread level filters to spread
    spread_filters = {f: filters[f] for f in filters if f.startswith("entry_spread")}
    return (
        _filter_data(spread, spread_filters)
        .pipe(_dedup_rows_by_cols, ["delta", "strike"])
        .sort_values(sort_by, ascending=ascending)
        .pipe(assign_trade_num, ["quote_date", "expiration", "underlying_symbol"])
    )


# this is the main function that runs the backtest engine
def run(data, trades, filters, init_balance=10000, mode="midpoint"):
    # for each option to be traded, determine the historical price action
    filters = {**default_entry_filters, **filters}
    exit_filters = {f: filters[f] for f in filters if f.startswith("exit_")}
    res = (
        pd.merge(trades, data, on=on, suffixes=("_entry", "_exit"))
        .pipe(_filter_data, exit_filters)
        .pipe(calc_entry_px, mode)
        .pipe(calc_exit_px, mode)
        .pipe(calc_pnl)
        .rename(columns=output_cols)
        .sort_values(["entry_date", "expiration", "underlying_symbol", "strike"])
        .pipe(assign_trade_num, ["entry_date", "expiration", "underlying_symbol"])
    )

    return calc_total_profit(res), res[output_format]



#     Optopsy - Python Backtesting library for options trading strategies
#     Copyright (C) 2018  Michael Chu

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

import glob
import os
import sys
import pandas as pd
from distutils.util import strtobool


# All recognized fields by the library are defined in the tuples below. Structs are used
# to map headers from source data to one of the recognized fields.
# The second item of each tuple defines if that field is required or not
# The third item of each tuple defines the expected value type of the field. This
# is used internally in the library and should not be changed.
# The fourth item of each tuple defines if the field is affected by ratios

fields = (
    ("option_symbol", False),
    ("underlying_symbol", True),
    ("quote_date", True),
    ("expiration", True),
    ("strike", True),
    ("option_type", True),
    ("bid", True),
    ("ask", True),
    ("underlying_price", True),
    ("implied_vol", False),
    ("delta", True),
    ("gamma", True),
    ("theta", True),
    ("vega", True),
    ("rho", False),
)


def _import_file(path, names, usecols, date_cols, skiprow):
    if not os.path.isdir(path):
        data = pd.read_csv(
            path,
            names=names,
            usecols=usecols,
            parse_dates=date_cols,
            skiprows=skiprow,
            infer_datetime_format=True,
        )
    elif os.path.isdir(path):
        fls = sorted(glob.glob(os.path.join(path, "*.csv")))
        data = pd.concat(
            pd.read_csv(
                f, names=names, usecols=usecols, parse_dates=date_cols, skiprows=skiprow
            )
            for f in fls
        )
    else:
        raise ValueError("Invalid path, please provide a valid path to a file")

    return data.pipe(format_option_df)


def _do_preview(data):
    print(data.head())
    return _user_prompt("Does this look correct?")


def get(file_path, struct, skiprow=1, prompt=True):
    return _do_import(file_path, struct, skiprow, prompt)


def _do_import(path, struct, skiprow, prompt):
    cols = list(zip(*struct))
    date_cols = [cols[0].index("quote_date"), cols[0].index("expiration")]

    if _check_structs(struct, cols):
        data = _import_file(path, cols[0], cols[1], date_cols, skiprow)
        if not prompt or (prompt & _do_preview(data)):
            return data
        else:
            sys.exit()


def format_option_df(df):
    return (
        df.assign(
            expiration=lambda r: pd.to_datetime(r["expiration"], format="%Y-%m-%d"),
            quote_date=lambda r: pd.to_datetime(r["quote_date"], format="%Y-%m-%d"),
            option_type=lambda r: r["option_type"].str.lower().str[:1],
        )
        .assign(dte=lambda r: (r["expiration"] - r["quote_date"]).dt.days)
        .round(2)
    )


def _check_field_is_standard(struct):
    # First we check if the provided struct uses our standard list of defined
    # column names
    std_fields = list(zip(*fields))[0]
    for f in struct:
        if f[0] not in std_fields or f[1] < 0:
            raise ValueError("Field names or field indices not valid!")
    return True


def _check_field_is_duplicated(cols):
    # check if we have any duplicate indices, which would be invalid
    if len(list(set(cols[1]))) != len(cols[1]):
        raise ValueError("Duplicate indices found!")
    return True


def _check_fields_contains_required(cols):
    # Check if the struct provided contains all the required fields
    req_fields = [x[0] for x in fields if x[1] is True]
    if not all(f in cols[0] for f in req_fields):
        raise ValueError("Required field names not defined!")
    return True


def _check_structs(struct, cols):
    return (
        _check_field_is_standard(struct)
        and _check_field_is_duplicated(cols)
        and _check_fields_contains_required(cols)
    )


def _user_prompt(question):
    while True:
        sys.stdout.write(question + " [y/n]: ")
        user_input = input().lower()
        try:
            result = strtobool(user_input)
            return result
        except ValueError:
            sys.stdout.write("Please use y/n or yes/no.\n")



#     Optopsy - Python Backtesting library for options trading strategies
#     Copyright (C) 2018  Michael Chu

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum


class Period(Enum):
    ONE_DAY = 1
    TWO_DAYS = 2
    THREE_DAYS = 3
    FOUR_DAYS = 4
    FIVE_DAYS = 5
    SIX_DAYS = 6
    SEVEN_DAYS = 7
    EIGHT_DAYS = 8
    NINE_DAYS = 9
    TEN_DAYS = 10
    ONE_WEEK = 7
    TWO_WEEKS = 14
    THREE_WEEKS = 21
    FOUR_WEEKS = 28
    FIVE_WEEKS = 35
    SIX_WEEKS = 42
    SEVEN_WEEKS = 49


class OptionType(Enum):
    CALL = ("c", 1)
    PUT = ("p", -1)


class OrderAction(Enum):
    BTO = (1, "BUY", "BOT")
    BTC = (1, "BUY", "BOT")
    STO = (-1, "SELL", "SLD")
    STC = (-1, "SELL", "SLD")


class DayOfWeek(Enum):
    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6



from setuptools import setup

setup(
    name="optopsy",
    description="Python Backtesting library for options trading strategies",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    version="1.0.3",
    url="https://github.com/michaelchu/optopsy",
    author="Michael Chu",
    author_email="mchchu88@gmail.com",
    license="GPL-3.0-or-later",
    classifiers=[
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3.6",
    ],
    packages=["optopsy"],
)



#     Optopsy - Python Backtesting library for options trading strategies
#     Copyright (C) 2018  Michael Chu

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from optopsy.data import fields
from optopsy.enums import Period, OptionType


def _convert(val):
    return val.value if isinstance(val, Period) else val


def _calc_abs_distance(row, column, val, absolute):
    col = abs(row[column]) if absolute else row[column]
    return abs(col - _convert(val))


def calls(df):
    return df[df.option_type.str.lower().str.startswith("c")]


def puts(df):
    return df[df.option_type.str.lower().str.startswith("p")]


def opt_type(df, option_type):
    if isinstance(option_type, OptionType):
        return df[df["option_type"] == option_type.value[0]]
    else:
        raise ValueError("option_type must be of type OptionType")


def underlying_price(df):
    if "underlying_price" in df:
        dates = df["underlying_price"].unique()
        return dates.mean()
    else:
        raise ValueError("Underlying Price column undefined!")


def nearest(df, column, val, groupby=None, absolute=True, tie="roundup"):
    # we need to group by unique option columns so that we are
    # getting the min abs dist over multiple sets of option groups
    # instead of the absolute min of the entire data set.
    if groupby is None:
        groupby = ["quote_date", "option_type", "expiration", "underlying_symbol"]

    on = groupby + ["abs_dist"]

    data = df.assign(abs_dist=lambda r: _calc_abs_distance(r, column, val, absolute))

    return (
        data.groupby(groupby)["abs_dist"]
        .min()
        .to_frame()
        .merge(data, on=on)
        .drop("abs_dist", axis=1)
    )


def lte(df, column, val):
    return df[df[column] <= _convert(val)]


def gte(df, column, val):
    return df[df[column] >= _convert(val)]


def eq(df, column, val):
    return df[df[column] == _convert(val)]


def lt(df, column, val):
    return df[df[column] < _convert(val)]


def gt(df, column, val):
    return df[df[column] > _convert(val)]


def ne(df, column, val):
    return df[df[column] != _convert(val)]


def between(df, column, start, end, inclusive=True, absolute=False):
    if absolute:
        temp_col = f"{column}_temp"
        df[temp_col] = abs(df[column])
    else:
        temp_col = column

    result = df[
        df[temp_col].between(_convert(start), _convert(end), inclusive=inclusive)
    ]
    return result.drop(temp_col, axis=1) if absolute else result



#     Optopsy - Python Backtesting library for options trading strategies
#     Copyright (C) 2018  Michael Chu

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .option_queries import between, nearest, eq
from datetime import datetime
import pandas as pd


def _process_tuples(data, col, groupby, value):
    if len(set(value)) == 1:
        return eq(data, col, value[1])
    else:
        return data.pipe(nearest, col, value[1], groupby=groupby).pipe(
            between, col, value[0], value[2], absolute=True
        )


def _process_values(data, col, value, groupby=None, valid_types=(int, float, tuple)):
    if not isinstance(value, valid_types):
        raise ValueError("Invalid value passed to filter")
    elif isinstance(value, tuple):
        return _process_tuples(data, col, groupby, value)
    else:
        return nearest(data, col, value, groupby=groupby)


def _calc_strike_pct(data, value, n, idx):
    if not isinstance(value, (int, float, tuple)):
        raise ValueError(f"Invalid value passed for leg {n+1} entry strike percentage")
    elif idx == n:
        return data.assign(
            strike_pct=lambda r: (r["strike"] / r["underlying_price"]).round(2)
        ).pipe(_process_values, "strike_pct", value)
    else:
        return data


def start_date(data, value, _idx):
    if isinstance(value, datetime):
        return data[data["expiration"] >= value]
    else:
        raise ValueError("Start Dates must of Date type")


def end_date(data, value, _idx):
    if isinstance(value, datetime):
        return data[data["expiration"] <= value]
    else:
        raise ValueError("End Dates must of Date type")


def std_expr(data, value, _idx):
    """
    Use all expirations, only standard or only non standard.

    For example, only using non standard would only use weekly,
    monthly and quarterly expirations.
    """
    return data


def contract_size(data, value, _idx):
    """
    Multiply the profit in the trades report by the Contract Size.

    For example, the profit is multiplied by 10 contract size.
    """
    if isinstance(value, int):
        return data.assign(contracts=value)
    else:
        raise ValueError("Contract sizes must of Int type")


def entry_dte(data, value, _idx):
    """
    Days to expiration min and max for the trade to be considered.

    For example, it will search options that have days to expiration
    between and including 20 to 55.
    """
    groupby = ["option_type", "expiration", "underlying_symbol"]
    return _process_values(data, "dte", value, groupby=groupby)


def entry_days(data, value, _idx):
    """
    Stagger trades every this many Entry Days.

    For example, there would be a new trade every 7 days from the last new trade.
    """
    pass


def leg1_delta(data, value, idx):
    """
    Absolute value of a delta of an option.
    """
    return _process_values(data, "delta", value) if idx == 0 else data


def leg2_delta(data, value, idx):
    """
    Absolute value of a delta of an option.
    """
    return _process_values(data, "delta", value) if idx == 1 else data


def leg3_delta(data, value, idx):
    """
    Absolute value of a delta of an option.
    """
    return _process_values(data, "delta", value) if idx == 2 else data


def leg4_delta(data, value, idx):
    """
    Absolute value of a delta of an option.
    """
    return _process_values(data, "delta", value) if idx == 3 else data


def leg1_strike_pct(data, value, idx):
    """
    Stock Percentage (strike / stock price).
    """
    return _calc_strike_pct(data, value, 0, idx)


def leg2_strike_pct(data, value, idx):
    """
    Stock Percentage (strike / stock price).
    """
    return _calc_strike_pct(data, value, 1, idx)


def leg3_strike_pct(data, value, idx):
    """
    Stock Percentage (strike / stock price).
    """
    return _calc_strike_pct(data, value, 2, idx)


def leg4_strike_pct(data, value, idx):
    """
    Stock Percentage (strike / stock price).
    """
    return _calc_strike_pct(data, value, 3, idx)


def entry_spread_price(data, value, _idx):
    """
    The net price of the spread.

    For example, it would set a min max of $0.10 to $0.20 and find only spreads with prices
    within that range.
    """
    pass


def entry_spread_delta(data, value, _idx):
    """
    The net delta of the spread.

    For example, it would set a min max of .30 to .40 and find only spreads with net deltas
    within that range.
    """
    pass


def entry_spread_yield(data, value, _idx):
    """
    Yield Percentage is (option entry price / stockprice).

    For example it would search options that have yldpct between
    and including .05 to .10 (5 and 10 percent).
    """
    pass


def exit_dte(data, value, _idx):
    """
    Exit the trade when the days to expiration left is equal to or below this.

    For example, it would exit a trade with 10 days to expiration.
    """
    if value is None:
        return data[data["quote_date_exit"] == data["expiration"]]
    else:
        groupby = ["option_type", "expiration", "underlying_symbol"]
        return _process_values(data, "dte_exit", value, groupby=groupby)


def exit_hold_days(data, value, _idx):
    """
    Exit the trade when the trade was held this many days.

    For example, it would exit a trade when the trade has been held for 20 days.
    """
    pass


def exit_leg_1_delta(data, value, idx):
    """
    Exit the trade when the delta of leg 1 is below the min or above the max.

    For example, it would exit when the delta of the
    first leg is below .10 or above .90 delta.
    """
    pass


def exit_leg_1_otm_pct(data, value, idx):
    """
    Exit the trade when the strike as a percent of stock price
    of leg 1 is below the min or above the max.

    For example, it would exit when the strike percentage
    of stock price is below 1.05 or above 1.20.
    """
    pass


def exit_profit_loss_pct(data, value, _idx):
    """
    Take profits and add stop loss to exit trade at these intervals.

    For example, set a stop loss of 50 (-50%) and a profit target at 200 (200%) on a long call.
    Set only a stop loss by leaving profit blank.
    """
    pass


def exit_spread_delta(data, value, _idx):
    """
    Exit and roll the trade if the spread total delta exceed the min or max value. For Example;

    spread delta = leg1ratio * leg1delta + leg2ratio * leg2delta
    """
    pass


def exit_spread_price(data, value, _idx):
    """
    Exit the trade when the trade price falls below the min or rises above the max.

    For example, it would exit if below 0.4 min or above $0.90 max price.
    """
    pass


def exit_strike_diff_pct(data, value, _idx):
    """
    Exit the trade when the trade price divided by the difference
    in strike prices falls below the min or rises above the max.

    For example, a $5 wide call spread would exit if the price of
    $1 divided by 5 is below the min of .20 or the price of 4.5
    that is above the max of .90.
    """
    pass
